package org.kobjects.jdbc.format.dbase;

import java.io.IOException;

import org.kobjects.util.GeneralComparator;


/**
 * @author haustein
 *         <p/>
 *         Manages an Index page including the offset of the pointer in the parent page
 *         <p/>
 *         The Page format is as follows:
 *         <p/>
 *         <table border="1">
 *         <tr>
 *         <td align="right">        00</td>
 *         <td valign="top"># Of valid entries (n); i=[0..n-1] for the record number and keys, i=[0..n] for the child nodes</td>
 *         </tr>
 *         <tr>
 *         <td align="right">i*krs + 04</td>
 *         <td valign="top">Index of left child node, 0 for leaf nodes</td></tr>
 *         <tr>
 *         <td align="right">i*krs + 08</td>
 *         <td valign="top">Record number, 0 if not leaf node.</td>
 *         </tr>
 *         <tr>
 *         <td align="right">i*krs + 12</td>
 *         <td>Key data</td>
 *         </tr>
 *         <tr></tr>s
 *         <tr>
 *         <td align="right">(i+1)*krs + 04</td>
 *         <td>child node (containing entries "between surrounding keys"), 0 for leaf nodes</td>
 *         </tr>
 *         <tr>
 *         <td align="right"> ...</td>
 *         <td>...</td>
 *         </tr>
 *         <tr></tr>
 *         <tr>
 *         <td align="right">n*krs + 04</td><td>"right" child node index</td>
 *         </tr>
 *         </table>
 */
public class NdxPage {

	byte[] pageData;
	int pageIndex;
	NdxFile ndx;
	NdxPage parentPage;
	int parentOfs;


	NdxPage(NdxFile ndx, int index, NdxPage parentPage, int parentOfs) {
		this.pageIndex = index;
		this.ndx = ndx;
		this.parentPage = parentPage;
		this.parentOfs = parentOfs;

		pageData = new byte[512 + ndx.keyRecordSize + 4]; // allow an additional record to simplify insert OPs

		ndx.load(index, pageData);

		//	System.out.println("page loaded: "+this);

	}

	NdxPage(NdxFile ndx, NdxPage parentPage, int parentOfs) throws IOException {
		this.ndx = ndx;
		this.pageIndex = ndx.getPageCount();
		this.parentPage = parentPage;
		this.parentOfs = parentOfs;

		pageData = new byte[512 + ndx.keyRecordSize + 4];

		ndx.setPageCount(pageIndex + 1);
		ndx.save();
	}


	int getCount() {
		return DbaseTable.getInt16(pageData, 0);
	}

	void setCount(int c) {
		DbaseTable.setInt16(pageData, 0, c);
	}

	int getRecNo(int i) {
		if (i <= 0 || i > getCount()) throw new IndexOutOfBoundsException();

		return DbaseTable.getInt32(pageData, 4 + (i - 1) * ndx.keyRecordSize + 4);
	}

	public void setRecNo(int i, int rnr) {
		if (i <= 0 || i > getCount()) throw new IndexOutOfBoundsException();

		DbaseTable.setInt32(pageData, 4 + (i - 1) * ndx.keyRecordSize + 4, rnr);
	}

	public int getPageIndex(int i) {
		if (i < 0 || i > getCount()) throw new IndexOutOfBoundsException();
		return DbaseTable.getInt32(pageData, i * ndx.keyRecordSize + 4);
	}

	public void setPageIndex(int i, int pi) {
		if (i < 0 || i > getCount()) throw new IndexOutOfBoundsException();
		DbaseTable.setInt32(pageData, i * ndx.keyRecordSize + 4, pi);
	}

	public NdxPage getPage(int i) {
		int pi = getPageIndex(i);
		return pi == 0 ? null : new NdxPage(ndx, pi, this, i);
	}

	public void setPage(int i, NdxPage p) {
		setPageIndex(i, p == null ? 0 : p.pageIndex);
	}


	/**
	 * Key indices range from 0 to getKeyCount()-1
	 */

	String getKey(int i) {
		if (i <= 0 || i > getCount())
			throw new IndexOutOfBoundsException();

		int st = 4 + (i - 1) * ndx.keyRecordSize + 8;
		return DbaseTable.getString(pageData, st, st + ndx.keyLength); //4+(i+1)*ndx.keyRecordSize+4);
	}

	public void setKey(int i, Object key) {
		if (i <= 0 || i > getCount())
			throw new IndexOutOfBoundsException();

		DbaseTable.setString(pageData, 4 + (i - 1) * ndx.keyRecordSize + 8, 4 + (i) * ndx.keyRecordSize, (String) key);
	}


	/**
	 *
	 *
	 *	 */


	void insertNode(int i, NdxPage node, Object key) throws IOException {
		System.out.println("Before insertNode(" + i + ", " + node.pageIndex + ", " + key + "): " + this);

//		System.arraycopy(pageData, 4+i*ndx.keyRecordSize+8, pageData, 4+(i+1)*ndx.keyRecordSize+8, (getCount()-i)*ndx.keyRecordSize-4);


		setCount(getCount() + 1);

		for (int j = getCount(); j > i; j--) {
			setKey(j, getKey(j - 1));
			setPageIndex(j, getPageIndex(j - 1));
		}

		setKey(i, key);
		setPage(i, node);

		System.out.println("After insertNode: " + this);

		if (getCount() >= ndx.keysPerPage) {
			System.out.println("Split...");

			NdxPage split = new NdxPage(ndx, parentPage, parentOfs + 1);
			int nc = getCount() / 2;
			int sc = getCount() - nc;
			split.setCount(sc);

			split.setPageIndex(0, getPageIndex(nc));
			for (int j = 1; j <= sc; j++) {
				split.setKey(j, getKey(nc + j));
				split.setPageIndex(j, getPageIndex(nc + j));
			}


			split.save();

			// insert largest key remaining in left child as delimiter
			parentPage.insertNode(parentOfs + 1, split, getKey(nc));

			setCount(nc - 1);

			System.out.println("Split to " + this + " and " + split);
		}

		save();

	}


	/**
	 * Insert a record in a leaf node. There are no pointers to handle.
	 */

	void insertRecord(int i, int rnr, Object key) throws IOException {

		System.out.println("Before insertRecord(" + i + ", " + rnr + ", " + key + "): " + this);

		setCount(getCount() + 1);

		setPageIndex(getCount(), 0);

		for (int j = getCount(); j > i; j--) {
			setRecNo(j, getRecNo(j - 1));
			setKey(j, getKey(j - 1));
		}

		setRecNo(i, rnr);
		setKey(i, key);

		System.out.println("After insertRecord: " + this);

		if (getCount() >= ndx.keysPerPage) {
			System.out.println("Split required!");

			NdxPage split = new NdxPage(ndx, parentPage, parentOfs + 1);
			int nc = getCount() / 2;
			int sc = getCount() - nc;
			split.setCount(sc);

			split.setPageIndex(0, 0);

			for (int j = 1; j <= sc; j++) {
				split.setKey(j, getKey(j + nc));
				split.setRecNo(j, getRecNo(j + nc));
				split.setPageIndex(j, 0);
			}

			split.save();

			// insert largest key remaining in left child as delimiter
			parentPage.insertNode(parentOfs + 1, split, getKey(nc));

			setCount(nc);
			System.out.println("Split to " + this + " and " + split);
		}

		save();
	}

	public void deleteRecord(int i) throws IOException {

		for (int j = i; j < getCount(); j++) {
			setRecNo(j, getRecNo(j + 1));
			setKey(j, getKey(j + 1));
		}

		setCount(getCount() - 1);

//		if(getCount() == 0)
		//		;//parentPage.delete(parentOfs);

		save();
	}


	public void save() throws IOException {
		ndx.raf.seek(pageIndex * 512);
		ndx.raf.write(pageData, 0, 512);
	}


	/**
	 * Should always return a leaf node. rnr is ignored if == 0
	 */

	NdxPos find(Object key, int rnr) {

		int index = 1;

		while (index <= getCount() && GeneralComparator.cmp(getKey(index), key) < 0) {
			index++;
		}

		if (getPage(index - 1) == null) {

			// leaf page found. now, what about rnr?!? 

			if (rnr != 0) {
				while (index <= getCount()
						&& rnr != getRecNo(index)
						&& GeneralComparator.cmp(key, getKey(index)) == 0) {


					index++;
				}
			}

			if (rnr != 0 && (index > getCount() || rnr != getRecNo(index))) {

				// this block is just for debug output! 
				NdxPage p = this;
				int pl = 0;
				while (p != null) {
					System.out.println((pl == 0 ? "self: " : ("parent" + pl + ": ")) + p);
					pl++;
					p = p.parentPage;
				}

				return null;
			}

			return new NdxPos(this, index);
		} else {
			index--;
			NdxPos result = null;
			do {
				result = getPage(index).find(key, rnr);
				index++;
			}
			while (result == null
					&& index <= getCount()
					&& GeneralComparator.cmp(key, getKey(index)) == 0);

			return result;
		}
	}


	public String toString() {
		StringBuffer buf = new StringBuffer("Page Nr. " + pageIndex
				+ " loaded; Parent Nr.: " + (parentPage != null ? parentPage.pageIndex : -1)
				+ " offset: " + parentOfs
				+ " getCount(): " + getCount());

		buf.append("\n - 0: " + getPageIndex(getCount()) + "\n");
		for (int i = 1; i <= getCount(); i++) {
			buf.append("\n - " + i + ": " + getPageIndex(i) + " / " + getRecNo(i) + " / " + getKey(i));
		}

		return buf.toString();
	}

}
