package org.kobjects.jdbc.ram;

import java.io.*;
import java.sql.*;
import java.util.*;

import org.kobjects.jdbc.*;
import org.kobjects.jdbc.sql.ConstantExpr;
import org.kobjects.jdbc.sql.Expression;
import org.kobjects.jdbc.stream.ReaderResultSet;
import org.kobjects.jdbc.stream.WriterResultSet;
import org.kobjects.jdbc.util.*;
import org.kobjects.util.Csv;

/**
 * Properties: readonly=true
 *
 * @author Stefan Haustein
 */

public class RamTable extends AbstractTable implements Runnable {

	private boolean inUpdate;
	protected String type;
	protected String file;
	protected Properties properties;

	Vector rows = new Vector();

	boolean modified;
	boolean leave;
	boolean readOnly;
	Thread thread;
	Column idColumn;
	static long lastId = 0;
	long updateInterval;
	long lastUpdate;
	Column updateColumn;


	public RamTable(String type, String file, Properties properties)
			throws ClassNotFoundException, SQLException {

		this.type = type;
		this.file = file;
		this.properties = properties;

		readOnly = "true".equalsIgnoreCase((String) properties.get("readonly"));

		ReaderResultSet source = getReaderResultSet();

		initStructure(source);
		initContent(source);

		if (!readOnly) {
			thread = new Thread(this);
			thread.start();
		}

		lastUpdate = System.currentTimeMillis();
	}

	protected ReaderResultSet getReaderResultSet() throws ClassNotFoundException, SQLException {
		return
				(ReaderResultSet) TableManager.getResultSet(
						type,
						file,
						properties,
						TableManager.READ);
	}


	/**
	 * Initializes the structure of the ram table from the source set with setSource
	 */

	protected void initStructure(ReaderResultSet source) throws SQLException {
		if (source == null)
			throw new NullPointerException("please call setReader prior to initStructure!");
		for (int i = 1; i <= source.getColumnSet().getColumnCount(); i++) {
			addColumn(new Column(source.getColumnSet().getColumn(i)));
		}

		if (properties.getProperty("autoid") != null) {
			String autoid = properties.getProperty("autoid");
			try {
				idColumn = getColumn(autoid);
			}
			catch (SQLException e) {
				idColumn = new Column(autoid, Types.VARCHAR);
				addColumn(idColumn);
			}
			addIndex(new RamIndex(this, new Order(idColumn, Order.DEFAULT, null)));
		}

		if (properties.getProperty("update") != null) {
			String[] update = Csv.decode(properties.getProperty("update"));
			updateColumn = getColumn(update[0]);
			updateInterval = Long.parseLong(update[1]);
		}

	}

	/**
	 * Initializes (reads) the content of the ram table from the source set with setSource.
	 * The source is closed at the end of this method
	 */

	protected void initContent(ReaderResultSet source) throws SQLException {
		Object[] row = new Object[getColumnCount()];
		int srcCols = source.getColumnSet().getColumnCount();
		while (source.next()) {
			for (int i = 0; i < srcCols; i++) {
				row[i] = source.getObject(i + 1);
			}
			addRow(row);
		}
		source.close();
		modified = false;
	}

	public synchronized void run() {
		while (!leave) {
			try {
				wait(60000);
			}
			catch (InterruptedException e) {
				;
			}

			if (modified) {

				try {
					rewrite();
				}
				catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
	}

	public boolean isModified() {
		return modified;
	}


	public synchronized void update() throws SQLException {
		inUpdate = true;
		boolean saveM = modified;

		try {

			ReaderResultSet rrs = getReaderResultSet();
			ResultSetMetaData rrsm = rrs.getMetaData();
			while (rrs.next()) {
				ResultSet trs = select(null, updateColumn.getName() + "=" + new ConstantExpr(rrs.getObject(updateColumn.getName())), null);
				boolean insert = !trs.next();
				if (insert) trs.moveToInsertRow();
				for (int i = 1; i < rrsm.getColumnCount(); i++) {
					trs.updateObject(rrsm.getColumnName(i), rrs.getObject(i));
				}
				if (insert) trs.insertRow();
				else trs.updateRow();
			}
		}
		catch (ClassNotFoundException e) {
			SQLException se = new SQLException(e.toString());
			se.initCause(e);
			throw se;
		}
		finally {
			modified = saveM;
			inUpdate = false;
			lastUpdate = System.currentTimeMillis();
		}


	}


	public void rewrite() throws IOException, ClassNotFoundException, SQLException {
		rewrite(new Iterator() {
			int i = 1;

			public Object next() {
				return new Integer(i++);
			}

			public boolean hasNext() {
				return i <= rows.size();
			}

			public void remove() {
			}
		});
	}

	public void addMissingColumns(ColumnSet columns) throws SQLException {
		for (int i = 1; i <= getColumnCount(); i++) {
			try {
				columns.findColumn(getColumnName(i));
			}
			catch (SQLException e) {
				columns.addColumn(new Column(getColumnName(i), getColumnType(i)));
			}
		}
	}


	protected void rewrite(Iterator rowIterator) throws IOException, ClassNotFoundException, SQLException {

		modified = false;

		File newFile = new File(file + ".new");
		OutputStream os = new FileOutputStream(newFile);
		WriterResultSet rs = getWriterResultSet(newFile.toString());

		addMissingColumns(rs.getColumnSet());

		while (rowIterator.hasNext()) {
			int i = ((Integer) rowIterator.next()).intValue();
			Object[] row = (Object[]) rows.elementAt(i - 1);
			if (row == null)
				continue;
			rs.moveToInsertRow();
			for (int j = 0; j < row.length; j++) {
				rs.updateObject(j + 1, row[j]);
			}
			rs.insertRow();
		}

		rs.close();

		File bakFile = new File(file + ".bak");
		bakFile.delete();
		new File(file).renameTo(bakFile);
		newFile.renameTo(new File(file));
	}


	public WriterResultSet getWriterResultSet(String newFile)
			throws IOException, ClassNotFoundException, SQLException {

		return (WriterResultSet) TableManager.getResultSet(
				type,
				newFile,
				properties,
				TableManager.WRITE);

//        return source.getWriter(os);
	}

	/**
	 * @see org.kobjects.jdbc.AbstractTable#set(int, int, java.lang.Object)
	 */
	public synchronized void setImpl(
			int row,
			Object[] value,
			boolean[] dirty) {

		Object[] rowBuf = ((Object[]) rows.elementAt(row - 1));

		for (int i = 0; i < value.length; i++) {
			if (dirty[i]) {
				rowBuf[i] = value[i];
				dirty[i] = false;
			}
		}

		modified = true;
	}

	/**
	 * @see org.kobjects.jdbc.AbstractTable#get(int, int)
	 */

	public Object get(int row, int column) {
		return ((Object[]) rows.elementAt(row - 1))[column - 1];
	}

	/**
	 * @see org.kobjects.jdbc.AbstractTable#getRowCount()
	 */

	public int getRowCount() {
		return rows.size();
	}

	public boolean rowDeleted(int row) {
		return rows.elementAt(row - 1) == null;
	}

	/**
	 * @see org.kobjects.jdbc.AbstractTable#addRow(java.lang.Object[])
	 */
	public synchronized void addRowImpl(Object[] row) {
		Object[] copy = new Object[row.length];
		System.arraycopy(row, 0, copy, 0, row.length);

		if (idColumn != null) {
			long id = System.currentTimeMillis();
			if (id <= lastId)
				id = lastId + 1;

			copy[idColumn.getNumber() - 1] = Long.toString(id, 36);

			lastId = id;
		}


		rows.addElement(copy);
		modified = true;

	}

	/**
	 * @see org.kobjects.jdbc.AbstractTable#deleteRow(int)
	 */
	public synchronized void deleteRowImpl(int row) {
		rows.setElementAt(null, row - 1);
		modified = true;
	}

	public void close() {
		if (thread != null) {
			synchronized (this) {
				leave = true;
				notify();
			}
			try {
				thread.join();
			}
			catch (InterruptedException e) {
			}
		}

	}


	public ResultSet select(int[] fields, Expression condition, Order orderBy) throws SQLException {
		if (updateColumn != null && System.currentTimeMillis() - lastUpdate > updateInterval && !inUpdate)
			update();

		return super.select(fields, condition, orderBy);
	}
}
