package org.kobjects.jdbc.sql;

import java.sql.SQLException;
import java.util.*;

/**
 * @author Stefan Haustein
 */
public class BooleanExpr extends Expression {

	public static final String[] OP_NAMES = {"AND", "OR", "NOT", "IS"};


	/**
	 * Type constant for AND operator. The value of a node of type AND is
	 * true if and only if all its children evaluate to true.
	 */
	public static final int AND = 0;

	/**
	 * Type constant for OR operator. The value of a node of type OR is
	 * true if at least one its children evaluates to true, and false otherwise.
	 */
	public static final int OR = 1;

	/**
	 * Type constant for XOR operator. A node of type XOR must have exactly two
	 * children. It evaluates to true if and only if its children evaluate
	 * to different truth values.
	 */
	public static final int IS = 2;


	int operator;
	Expression[] children;

	public BooleanExpr(int operator, Vector children) {
		this.operator = operator;
		this.children = new Expression[children.size()];
		for (int i = 0; i < children.size(); i++) {
			this.children[i] = (Expression) children.elementAt(i);
		}
	}

	public Object eval(int rowNumber, Object[] rowData) throws SQLException {
		switch (operator) {
			case AND: {
				Boolean tmp = Boolean.TRUE;

				for (int i = 0; i < children.length; i++) {
					Boolean b = (Boolean) children[i].eval(rowNumber, rowData);
					if (b == null)
						tmp = null;
					else if (!b.booleanValue()) return Boolean.FALSE;
				}

				return tmp;
			}

			case OR: {
				Boolean tmp = Boolean.FALSE;

				for (int i = 0; i < children.length; i++) {
					Boolean b = (Boolean) children[i].eval(rowNumber, rowData);
					if (b == null)
						tmp = null;
					else if (b.booleanValue()) return Boolean.TRUE;
				}

				return tmp;
			}

			case IS: {
				Boolean bo = (Boolean) children[0].eval(rowNumber, rowData);
				if (bo == null) return null;
				boolean b = bo.booleanValue();
				for (int i = 1; i < children.length; i++) {
					bo = (Boolean) children[i].eval(rowNumber, rowData);
					if (bo == null) return null;
					b = b == bo.booleanValue();
				}
				return b ? Boolean.TRUE : Boolean.FALSE;
			}
/*
               case NOT: {
                   return ((Boolean)children[0].eval(row)).booleanValue()
                        ? Boolean.FALSE : Boolean.TRUE;
               }*/
		}
		throw new RuntimeException("Illegal op:" + operator);
	}


	public void getRangeRestrictions(Hashtable result) {
		if (operator != AND) return;

		for (int i = 0; i < children.length; i++) {
			children[i].getRangeRestrictions(result);
		}
	}

	public String toString(int sqlType) {
		StringBuffer buf = new StringBuffer("( ");
		buf.append(children[0].toString(sqlType));
		for (int i = 1; i < children.length; i++) {
			buf.append(' ');
			buf.append(OP_NAMES[operator]);
			buf.append(' ');
			buf.append(children[i].toString(sqlType));
		}
		buf.append(" ) ");
		return buf.toString();
	}


	public Class getType() {
		return Boolean.class;
	}
}
