package org.kobjects.jdbc.stream;

import java.io.*;
import java.net.HttpURLConnection;
import java.net.URL;
import java.sql.SQLException;
import java.util.Properties;


import org.kobjects.jdbc.util.AbstractResultSet;

/**
 * Intended to simplify ResultSet access to sequential file reading.
 * Implementations must provide a public constructor with a String
 * parameter denoting the source file and a Properties parameter.
 */

public abstract class ReaderResultSet
		extends AbstractResultSet {

	Object[] currentRow;
	Object[] peek;
	int row;
	boolean wasNull;


	protected abstract Object[] readRow()
			throws SQLException, IOException;

	/**
	 * Utility method performing some filename normalization
	 */

	public static Reader getReader(String fileName, Properties properties) throws IOException {
		String enc = properties.getProperty("encoding");

		if (fileName.startsWith("http://")) {
			HttpURLConnection hc = (HttpURLConnection) new URL(fileName).openConnection();
			String e2 = hc.getContentEncoding();
			if (e2 != null) enc = e2;

			return enc == null
					? new InputStreamReader(hc.getInputStream())
					: new InputStreamReader(hc.getInputStream(), enc);
		}

		if (fileName.startsWith("file:///"))
			fileName = fileName.substring(7);

		return enc == null ? new FileReader(fileName) : new InputStreamReader(new FileInputStream(fileName), enc);
	}


	/**
	 * @see java.sql.ResultSet#next()
	 */
	public boolean next() throws SQLException {

		try {
			if (peek != null) {
				currentRow = peek;
				peek = null;
			} else currentRow = readRow();
		}
		catch (IOException e) {
			throw new SQLException(e.toString());
		}
		row++;

		return currentRow != null;
	}
	/*
		 * @see java.sql.ResultSet#close()
		public void close() throws SQLException {
		}

			 */

	/**
	 * @see java.sql.ResultSet#wasNull()
	 */
	public boolean wasNull() throws SQLException {
		return wasNull;
	}

	/**
	 * @see java.sql.ResultSet#getObject(int)
	 */
	public Object getObject(int columnIndex)
			throws SQLException {
		if (currentRow == null)
			throw new SQLException("get before first or after last");

		Object result = currentRow[columnIndex - 1];
		wasNull = result == null;
		return result;
	}

	/**
	 * @see java.sql.ResultSet#isAfterLast()
	 */
	public boolean isAfterLast() throws SQLException {
		return currentRow == null && row != 0;
	}

	/**
	 * @see java.sql.ResultSet#isLast()
	 */
	public boolean isLast() throws SQLException {
		if (currentRow == null) return false;
		if (peek != null) return true;

		try {
			peek = readRow();
		}
		catch (IOException e) {
			throw new SQLException(e.toString());
		}

		return peek != null;
	}

	/**
	 * @see java.sql.ResultSet#getRow()
	 */
	public int getRow() throws SQLException {
		return row;
	}

	/**
	 * @see java.sql.ResultSet#absolute(int)
	 */
	public boolean absolute(int row) throws SQLException {
		throw new SQLException("Only sequential read allowed");
	}

	/**
	 * @see java.sql.ResultSet#rowUpdated()
	 */
	public boolean rowUpdated() throws SQLException {
		return false;
	}

	/**
	 * @see java.sql.ResultSet#rowInserted()
	 */
	public boolean rowInserted() throws SQLException {
		return false;
	}

	/**
	 * @see java.sql.ResultSet#rowDeleted()
	 */
	public boolean rowDeleted() throws SQLException {
		return false;
	}

	/**
	 * @see java.sql.ResultSet#updateObject(int, java.lang.Object)
	 */
	public void updateObject(int columnIndex, Object x)
			throws SQLException {
		throw new SQLException("This ResultSet is read-only!");
	}

	/**
	 * @see java.sql.ResultSet#insertRow()
	 */
	public void insertRow() throws SQLException {
		throw new SQLException("This ResultSet is read-only!");
	}

	/**
	 * @see java.sql.ResultSet#updateRow()
	 */
	public void updateRow() throws SQLException {
		throw new SQLException("This ResultSet is read-only!");
	}

	/**
	 * @see java.sql.ResultSet#deleteRow()
	 */
	public void deleteRow() throws SQLException {
		throw new SQLException("This ResultSet is read-only!");
	}

	/**
	 * @see java.sql.ResultSet#refreshRow()
	 */
	public void refreshRow() throws SQLException {
	}

	/**
	 * @see java.sql.ResultSet#moveToInsertRow()
	 */
	public void moveToInsertRow() throws SQLException {
		throw new SQLException("This ResultSet is read-only!");
	}

	/**
	 * @see java.sql.ResultSet#moveToCurrentRow()
	 */
	public void moveToCurrentRow() throws SQLException {
	}

}
