package org.kobjects.jdbc.util;

import java.io.*;
import java.sql.*;
import java.util.*;

import org.kobjects.jdbc.*;
import org.kobjects.jdbc.sql.Expression;
import org.kobjects.util.GeneralComparator;

/**
 * @author Stefan Haustein
 */

public class TableResultSet extends AbstractResultSet {

	AbstractTable table;
	int row;
	Object[] writeBuf;
	boolean wasNull;
	boolean[] dirty;
	boolean insertRow;
	Vector rowMap = new Vector();

	int[] fieldMap;
	Expression condition;
	static Comparator comparator = GeneralComparator.getInstance();

	/**
	 * Constructor TableResultSet.
	 *
	 * @param table
	 */

	public TableResultSet(AbstractTable table, int[] fieldMap, Expression condition, Order orderBy)
			throws SQLException {
		this.table = table;

		if (fieldMap == null) {
			fieldMap = new int[table.getColumnCount()];
			for (int i = 0; i < fieldMap.length; i++)
				fieldMap[i] = i + 1;
			this.fieldMap = fieldMap;
		} else {
			this.fieldMap = new int[fieldMap.length];
			System.arraycopy(fieldMap, 0, this.fieldMap, 0, fieldMap.length);
		}

		for (int i = 0; i < fieldMap.length; i++) {
			getColumnSet().addColumn(new Column(table.getColumn(fieldMap[i])));
		}

		// check whether an index can be used to check (parts of) condition

		int bestSupport = 0;
		Index rangeIndex = null;
		Hashtable ranges = null;

		if (condition != null) {

			ranges = new Hashtable();
			condition.getRangeRestrictions(ranges);

			for (int i = 0; i < table.getIndexCount(); i++) {
				int support = table.getIndex(i).getOrder().getRangeSupport(ranges);
				if (support > bestSupport) {
					rangeIndex = table.getIndex(i);
					bestSupport = support;
				}
			}
		}

		if (bestSupport > 0) {
			System.out.println("query optimization; using index: " + rangeIndex);

			Object[] mins = new Object[bestSupport];
			Comparable[] maxs = new Comparable[bestSupport];
			int[] cols = new int[bestSupport];

			Order order = rangeIndex.getOrder();
			for (int i = 0; i < bestSupport; i++) {
				Column c = order.getColumn();
				Object[] range = (Object[]) ranges.get(c);
				mins[i] = range[0];
				maxs[i] = (Comparable) range[1];
				cols[i] = c.getNumber();
				order = order.getRemainder();
			}

			Iterator i = rangeIndex.getIterator(mins);
			boolean exceeded = false;
			while (i.hasNext()) {
				// && ((Comparable) i.getKey()).compareTo(range[1]) <= 0) {
				int ri = ((Integer) i.next()).intValue();
				if (table.rowDeleted(ri)) continue;
				for (int j = 0; j < bestSupport; j++) {
					System.out.println("comparing " + maxs[j] + " to " + table.get(ri, cols[j]));
					if (maxs[j] != null && comparator.compare(
							maxs[j],
							table.get(ri, cols[j])) < 0) {
						System.out.println("failed, leaving!");
						exceeded = true;
						break;
					}
				}
				if (exceeded) break;

				if (((Boolean) condition.eval(ri, null)).booleanValue())
					rowMap.addElement(new Integer(ri));
			}
		} else {
			for (int i = 1; i <= table.getRowCount(); i++) {
				if (!table.rowDeleted(i)) {
					if (condition == null)
						rowMap.addElement(new Integer(i));
					else {
						Boolean b = (Boolean) condition.eval(i, null);
						if (b != null && b.booleanValue())
							rowMap.addElement(new Integer(i));
					}
				}
			}
		}

		if (orderBy != null) {
			Collections.sort(rowMap, orderBy);
		}
	}

	/**
	 * @see java.sql.ResultSet#next()
	 */

	public boolean next() throws SQLException {
		return absolute(row + 1);
	}

	/**
	 * @see java.sql.ResultSet#close()
	 */

	public void close() throws SQLException {
	}

	/**
	 * @see java.sql.ResultSet#wasNull()
	 */
	public boolean wasNull() throws SQLException {
		return wasNull;
	}

	int getTableColumn(int column) {
		return fieldMap[column - 1];
	}

	int getTableRow(int row) {
		return ((Integer) rowMap.elementAt(row - 1)).intValue();
	}

	/**
	 * @see java.sql.ResultSet#getObject(int)
	 */
	public Object getObject(int columnIndex) throws SQLException {
		int tc = getTableColumn(columnIndex);

		Object result =
				writeBuf != null
						&& (insertRow || dirty[tc - 1])
						? writeBuf[tc - 1]
						: table.get(getTableRow(row), tc);

		wasNull = result == null;
		return result;
	}

	/**
	 * @see java.sql.ResultSet#isAfterLast()
	 */
	public boolean isAfterLast() throws SQLException {
		return row > rowMap.size();
	}

	/**
	 * @see java.sql.ResultSet#isLast()
	 */
	public boolean isLast() throws SQLException {
		return row == rowMap.size() && row != 0;
	}

	/**
	 * @see java.sql.ResultSet#getRow()
	 */
	public int getRow() throws SQLException {
		return row;
	}

	/**
	 * @see java.sql.ResultSet#absolute(int)
	 */
	public boolean absolute(int row) throws SQLException {
		writeBuf = null;
		if (row > table.getRowCount() + 1)
			return false;
		this.row = row;
		return row <= rowMap.size();
	}

	/**
	 * @see java.sql.ResultSet#rowUpdated()
	 */
	public boolean rowUpdated() throws SQLException {
		throw new RuntimeException("NYI");
	}

	/**
	 * @see java.sql.ResultSet#rowInserted()
	 */
	public boolean rowInserted() throws SQLException {
		throw new RuntimeException("NYI");
	}

	/**
	 * @see java.sql.ResultSet#rowDeleted()
	 */
	public boolean rowDeleted() throws SQLException {
		return table.rowDeleted(getTableRow(row));
	}

	/**
	 * @see java.sql.ResultSet#updateObject(int, java.lang.Object)
	 */
	public void updateObject(int columnIndex, Object x) throws SQLException {

		if (writeBuf == null) {
			writeBuf = new Object[table.getColumnCount()];
			dirty = new boolean[table.getColumnCount()];
		}

		writeBuf[getTableColumn(columnIndex) - 1] = x;
		dirty[getTableColumn(columnIndex) - 1] = true;
	}


	public void updateBinaryStream(int columnIndex, InputStream x, int length)
			throws SQLException {

		try {
			Blob blob = (Blob) getObject(columnIndex);

			if (blob == null)
				blob = (Blob) table.createLob(columnIndex, insertRow ? -1 : getTableRow(row));

			OutputStream out = blob.setBinaryStream(1);
			byte[] buf = new byte[1024];

			int remaining = length;
			while (remaining > 0) {
				int cnt = x.read(buf, 0, Math.min(remaining, buf.length));
				out.write(buf, 0, cnt);
				remaining -= cnt;
			}

			out.close();

			blob.truncate(length);
		}
		catch (IOException e) {
			throw new SQLException(e.toString());
		}
	}

	/**
	 * @see java.sql.ResultSet#insertRow()
	 */
	public void insertRow() throws SQLException {
		if (!insertRow)
			throw new SQLException("Not on insert row!");

		table.addRow(writeBuf);
	}

	/**
	 * @see java.sql.ResultSet#updateRow()
	 */
	public void updateRow() throws SQLException {
		if (writeBuf == null)
			return;

		if (insertRow || row <= 0 || row > rowMap.size())
			throw new SQLException("Illegal row index for update:" + row);

		table.set(getTableRow(row), writeBuf, dirty);

		/*		for (int i = 0; i < writeBuf.length; i++) {
				    if(dirty[i])
				    	table.set(row, i+1, writeBuf[i]);
				    	
				    dirty[i] = false;
				}*/
	}

	/**
	 * @see java.sql.ResultSet#deleteRow()
	 */

	public void deleteRow() throws SQLException {
		table.deleteRow(getTableRow(row));
		//        throw new RuntimeException("NYI");
	}

	/**
	 * @see java.sql.ResultSet#refreshRow()
	 */
	public void refreshRow() throws SQLException {
		writeBuf = null;
	}

	/**
	 * @see java.sql.ResultSet#moveToInsertRow()
	 */
	public void moveToInsertRow() throws SQLException {
		insertRow = true;
		writeBuf = new Object[table.getColumnCount()];
		dirty = new boolean[table.getColumnCount()];
	}

	/**
	 * @see java.sql.ResultSet#moveToCurrentRow()
	 */
	public void moveToCurrentRow() throws SQLException {
		if (insertRow) {
			insertRow = false;
			writeBuf = null;
		}

	}

	public RowId getRowId(int columnIndex) throws SQLException {
		return null;
	}

	public RowId getRowId(String columnLabel) throws SQLException {
		return null;
	}

	public void updateRowId(int columnIndex, RowId x) throws SQLException {
	}

	public void updateRowId(String columnLabel, RowId x) throws SQLException {
	}

	public int getHoldability() throws SQLException {
		return 0;
	}

	public boolean isClosed() throws SQLException {
		return false;
	}

	public void updateNString(int columnIndex, String nString) throws SQLException {
	}

	public void updateNString(String columnLabel, String nString) throws SQLException {
	}

	public void updateNClob(int columnIndex, NClob nClob) throws SQLException {
	}

	public void updateNClob(String columnLabel, NClob nClob) throws SQLException {
	}

	public NClob getNClob(int columnIndex) throws SQLException {
		return null;
	}

	public NClob getNClob(String columnLabel) throws SQLException {
		return null;
	}

	public SQLXML getSQLXML(int columnIndex) throws SQLException {
		return null;
	}

	public SQLXML getSQLXML(String columnLabel) throws SQLException {
		return null;
	}

	public void updateSQLXML(int columnIndex, SQLXML xmlObject) throws SQLException {
	}

	public void updateSQLXML(String columnLabel, SQLXML xmlObject) throws SQLException {
	}

	public String getNString(int columnIndex) throws SQLException {
		return null;
	}

	public String getNString(String columnLabel) throws SQLException {
		return null;
	}

	public Reader getNCharacterStream(int columnIndex) throws SQLException {
		return null;
	}

	public Reader getNCharacterStream(String columnLabel) throws SQLException {
		return null;
	}

	public void updateNCharacterStream(int columnIndex, Reader x, long length) throws SQLException {
	}

	public void updateNCharacterStream(String columnLabel, Reader reader, long length) throws SQLException {
	}

	public void updateAsciiStream(int columnIndex, InputStream x, long length) throws SQLException {
	}

	public void updateBinaryStream(int columnIndex, InputStream x, long length) throws SQLException {
	}

	public void updateCharacterStream(int columnIndex, Reader x, long length) throws SQLException {
	}

	public void updateAsciiStream(String columnLabel, InputStream x, long length) throws SQLException {
	}

	public void updateBinaryStream(String columnLabel, InputStream x, long length) throws SQLException {
	}

	public void updateCharacterStream(String columnLabel, Reader reader, long length) throws SQLException {
	}

	public void updateBlob(int columnIndex, InputStream inputStream, long length) throws SQLException {
	}

	public void updateBlob(String columnLabel, InputStream inputStream, long length) throws SQLException {
	}

	public void updateClob(int columnIndex, Reader reader, long length) throws SQLException {
	}

	public void updateClob(String columnLabel, Reader reader, long length) throws SQLException {
	}

	public void updateNClob(int columnIndex, Reader reader, long length) throws SQLException {
	}

	public void updateNClob(String columnLabel, Reader reader, long length) throws SQLException {
	}

	public void updateNCharacterStream(int columnIndex, Reader x) throws SQLException {
	}

	public void updateNCharacterStream(String columnLabel, Reader reader) throws SQLException {
	}

	public void updateAsciiStream(int columnIndex, InputStream x) throws SQLException {
	}

	public void updateBinaryStream(int columnIndex, InputStream x) throws SQLException {
	}

	public void updateCharacterStream(int columnIndex, Reader x) throws SQLException {
	}

	public void updateAsciiStream(String columnLabel, InputStream x) throws SQLException {
	}

	public void updateBinaryStream(String columnLabel, InputStream x) throws SQLException {
	}

	public void updateCharacterStream(String columnLabel, Reader reader) throws SQLException {
	}

	public void updateBlob(int columnIndex, InputStream inputStream) throws SQLException {
	}

	public void updateBlob(String columnLabel, InputStream inputStream) throws SQLException {
	}

	public void updateClob(int columnIndex, Reader reader) throws SQLException {
	}

	public void updateClob(String columnLabel, Reader reader) throws SQLException {
	}

	public void updateNClob(int columnIndex, Reader reader) throws SQLException {
	}

	public void updateNClob(String columnLabel, Reader reader) throws SQLException {
	}

	public <T> T unwrap(Class<T> iface) throws SQLException {
		return null;
	}

	public boolean isWrapperFor(Class<?> iface) throws SQLException {
		return false;
	}
}
